"""
Zemfrog is a simple framework based on flask for building a REST API quickly.
Which focuses on building a customizable, flexible and manageable REST API!
"""

__author__ = "Aprila Hijriyan"
__email__ = "hijriyan23@gmail.com"
__version__ = "5.0.1"



from zemfrog.globals import db
from zemfrog.mixins import (
    LogMixin,
    PermissionLinksMixin,
    PermissionMixin,
    RoleLinksMixin,
    RoleMixin,
    UserMixin,
)


class User(UserMixin, db.Model):
    pass


class Role(RoleMixin, db.Model):
    pass


class RoleLinks(RoleLinksMixin, db.Model):
    pass


class Permission(PermissionMixin, db.Model):
    pass


class PermissionLinks(PermissionLinksMixin, db.Model):
    pass


class Log(LogMixin, db.Model):
    pass



#!/usr/bin/env python

"""Tests for `zemfrog` package."""


import unittest

from click.testing import CliRunner

from zemfrog import cli


class TestZemfrog(unittest.TestCase):
    """Tests for `zemfrog` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_something(self):
        """Test something."""

    def test_command_line_interface(self):
        """Test the CLI."""
        runner = CliRunner()
        result = runner.invoke(cli.main)
        assert result.exit_code == 0



from flask import jsonify
from zemfrog.models import DefaultResponseSchema
from zemfrog.decorators import marshal_with, http_code


@marshal_with(200, DefaultResponseSchema)
@http_code
def get():
    code = 200
    message = "What's up?"
    return {"code": code, "message": message}


tag = "{{ name }}"
description = "API"
url_prefix = "/{{url_prefix | lower}}"
routes = [("/get", get, ["GET"])]



from flask import jsonify


def handler(error):
    """
    Reference: https://webargs.readthedocs.io/en/latest/framework_support.html#error-handling
    """

    headers = error.data.get("headers", None)
    messages = error.data.get("messages", ["Invalid request."])
    if headers:
        return jsonify({"errors": messages}), error.code, headers
    else:
        return jsonify({"errors": messages}), error.code



from flask import current_app
from flask_marshmallow import EXTENSION_NAME
from werkzeug.local import LocalProxy


def _get_db():
    db = current_app.extensions["sqlalchemy"].db
    return db


def _get_smorest():
    api = current_app.extensions["flask-smorest"]["ext_obj"]
    return api


def _get_celery():
    celery = current_app.extensions["celery"]
    return celery


def _get_mail():
    mail = current_app.extensions["mail"]
    return mail


def _get_marshmallow():
    ma = current_app.extensions[EXTENSION_NAME]
    return ma


def _get_migrate():
    migrate = current_app.extensions["migrate"]
    return migrate


db = LocalProxy(_get_db)
smorest = LocalProxy(_get_smorest)
celery = LocalProxy(_get_celery)
mail = LocalProxy(_get_mail)
ma = LocalProxy(_get_marshmallow)
migrate = LocalProxy(_get_migrate)



from flask_smorest import Blueprint

def init_blueprint():
    blueprint = Blueprint(
        "{{name}}",
        __name__,
        url_prefix="/{{name}}",
        static_folder="static",
        template_folder="templates",
    )
    return blueprint



import os

from flask.globals import current_app
from jinja2 import Template

from .helper import copy_template, get_import_name, get_template, search_model
from .validators import validate_module_name


def g_project(name: str, import_name: str):
    """
    Functions for creating projects.

    :param name: project name.

    """

    validate_module_name(name)
    if os.path.isdir(name):
        print("The %r project already exists" % name)
        return

    print("Creating %r project... " % name, end="")
    copy_template("project", name)
    main_app = True if import_name == "wsgi" else False
    for root, _, files in os.walk(name):
        for f in files:
            if not f.endswith((".py", ".rst")):
                continue

            f = os.path.join(root, f)
            with open(f) as fp:
                data = fp.read()
                t = Template(data)
                new = t.render(import_name=import_name, main_app=main_app, name=name)

            with open(f, "w") as fp:
                fp.write(new)

    print("(done)")


def g_extension(name: str):
    """
    Function for making extension boilerplate.

    :param name: extension name.
    """

    validate_module_name(name)
    ext_dir = os.path.join(current_app.root_path, "extensions")
    old_filename = get_template("extension", "name.py")
    with open(old_filename) as fp:
        new_data = fp.read()

    new_filename = os.path.join(ext_dir, name + ".py")
    if os.path.isfile(new_filename):
        print("The %r extension already exists" % name)
        return

    print("Creating a %r extension..." % name, end="")
    with open(new_filename, "w") as fp:
        fp.write(new_data)

    print("(done)")


def g_task(name: str):
    """
    Function for creating celery task.

    :param name: task name.
    """

    validate_module_name(name)
    import_name = get_import_name(current_app)
    main_app = True
    if import_name:
        main_app = False

    ext_dir = os.path.join(current_app.root_path, "tasks")
    old_filename = get_template("task", "name.py")
    with open(old_filename) as fp:
        old_data = fp.read()
        py_t = Template(old_data)
        new_data = py_t.render(main_app=main_app, task_name=name)

    new_filename = os.path.join(ext_dir, name + ".py")
    if os.path.isfile(new_filename):
        print("The %r task already exists" % name)
        return

    print("Creating %r task..." % name, end="")
    with open(new_filename, "w") as fp:
        fp.write(new_data)

    print("(done)")


def g_model(name: str):
    """
    Function for creating sqlalchemy model.

    :param name: model name.
    """

    validate_module_name(name)
    import_name = get_import_name(current_app)
    main_app = True
    if import_name:
        main_app = False

    model_dir = os.path.join(current_app.root_path, "models")
    old_filename = get_template("model", "name.py")
    with open(old_filename) as fp:
        old_data = fp.read()
        py_t = Template(old_data)
        new_data = py_t.render(main_app=main_app, model_name=name)

    new_filename = os.path.join(model_dir, name + ".py")
    if os.path.isfile(new_filename):
        print("The %r model already exists" % name)
        return

    print("Creating %r model..." % name, end="")
    with open(new_filename, "w") as fp:
        fp.write(new_data)

    print("(done)")


def g_api(name: str):
    """
    Functions for creating APIs.

    :param name: REST API name.

    """

    validate_module_name(name)
    api_dir = os.path.join(current_app.root_path, "apis")
    old_filename = get_template("api", "name.py")
    with open(old_filename) as fp:
        old_data = fp.read()
        py_t = Template(old_data)
        new_data = py_t.render(name=name, url_prefix=name)

    new_filename = os.path.join(api_dir, name + ".py")
    if os.path.isfile(new_filename):
        print("The %r REST API already exists" % name)
        return

    print("Creating REST API %r... " % name, end="")
    with open(new_filename, "w") as fp:
        fp.write(new_data)

    print("(done)")


def g_api_crud(name: str):
    """
    Function for creating REST API.

    :param name: REST API name.

    """

    validate_module_name(name)
    src_model = search_model(name)
    import_name = get_import_name(current_app)
    main_app = True
    if import_name:
        main_app = False
        idx = len(import_name)
        src_model = src_model[idx:]

    api_dir = os.path.join(current_app.root_path, "apis")
    old_filename = get_template("crud", "name.py")
    with open(old_filename) as fp:
        old_data = fp.read()
        py_t = Template(old_data)
        new_data = py_t.render(
            name=name, url_prefix=name, src_model=src_model, main_app=main_app
        )

    new_filename = os.path.join(api_dir, name + ".py")
    if os.path.isfile(new_filename):
        print("The %r REST API already exists" % name)
        return

    print("Creating REST API %r... " % name, end="")
    with open(new_filename, "w") as fp:
        fp.write(new_data)

    print("(done)")


def g_blueprint(name: str):
    """
    Function for creating blueprints.

    :param name: blueprint name.

    """

    validate_module_name(name)
    bp_dir = os.path.join(current_app.root_path, name)
    if os.path.isdir(bp_dir):
        print("The %r blueprint already exists" % name)
        return

    print("Creating blueprint %r... " % name, end="")
    copy_template("blueprint", bp_dir)
    for fname in ("routes", "urls"):
        filename = os.path.join(bp_dir, fname + ".py")
        with open(filename) as fp:
            old_data = fp.read()
            py_t = Template(old_data)
            new_data = py_t.render(name=name)

        with open(filename, "w") as fp:
            fp.write(new_data)

    print("(done)")


def g_middleware(name: str):
    """
    Function for creating middleware.

    :param str name: middleware name.

    """

    validate_module_name(name)
    middleware_dir = os.path.join(current_app.root_path, "middlewares")
    old_filename = get_template("middleware", "name.py")
    with open(old_filename) as fp:
        data = fp.read()

    new_filename = os.path.join(middleware_dir, name + ".py")
    if os.path.isfile(new_filename):
        print("The %r middleware already exists" % name)
        return

    print("Creating middleware %r... " % name, end="")
    with open(new_filename, "w") as fp:
        fp.write(data)

    print("(done)")


def g_command(name: str):
    """
    Function to create boilerplate command.

    :param name: command name.

    """

    validate_module_name(name)
    cmd_dir = os.path.join(current_app.root_path, "commands")
    old_filename = get_template("command", "name.py")
    with open(old_filename) as fp:
        old_data = fp.read()
        py_t = Template(old_data)
        new_data = py_t.render(name=name)

    new_filename = os.path.join(cmd_dir, name + ".py")
    if os.path.isfile(new_filename):
        print("The %r command already exists" % name)
        return

    print("Creating command %r..." % name, end="")
    with open(new_filename, "w") as fp:
        fp.write(new_data)

    print("(done)")


def g_error_handler(name):
    """
    Function for creating error handler.

    :param str name: name of the handler.

    """

    validate_module_name(name)
    eh_dir = os.path.join(current_app.root_path, "handlers")
    old_filename = get_template("errorhandler", "name.py")
    with open(old_filename) as fp:
        data = fp.read()

    new_filename = os.path.join(eh_dir, name + ".py")
    if os.path.isfile(new_filename):
        print("The %r error handler already exists" % name)
        return

    print("Creating error handler %r... " % name, end="")
    with open(new_filename, "w") as fp:
        fp.write(data)

    print("(done)")


def g_loader(name):
    """
    The function to create our loader.

    :param str name: loader name.

    """

    validate_module_name(name)
    ld_dir = os.path.join(current_app.root_path, "loaders")
    old_filename = get_template("loader", "name.py")
    with open(old_filename) as fp:
        data = fp.read()

    new_filename = os.path.join(ld_dir, name + ".py")
    if os.path.isfile(new_filename):
        print("The %r loader already exists" % name)
        return

    print("Creating a loader %r... " % name, end="")
    with open(new_filename, "w") as fp:
        fp.write(data)

    print("(done)")


def g_context(name):
    validate_module_name(name)
    ct_dir = os.path.join(current_app.root_path, "contexts")
    old_filename = get_template("context", "name.py")
    with open(old_filename) as fp:
        data = fp.read()

    new_filename = os.path.join(ct_dir, name + ".py")
    if os.path.isfile(new_filename):
        print("The %r context already exists" % name)
        return

    print("Creating a context %r... " % name, end="")
    with open(new_filename, "w") as fp:
        fp.write(data)

    print("(done)")


def g_filter(name):
    validate_module_name(name)
    filter_dir = os.path.join(current_app.root_path, "filters")
    old_filename = get_template("filter", "name.py")
    with open(old_filename) as fp:
        data = fp.read()

    new_filename = os.path.join(filter_dir, name + ".py")
    if os.path.isfile(new_filename):
        print("The %r filter already exists" % name)
        return

    print("Creating a jinja filter %r... " % name, end="")
    with open(new_filename, "w") as fp:
        fp.write(data)

    print("(done)")



from flask import Flask
from flask_smorest import Blueprint

from ..globals import smorest
from ..helper import get_import_name, import_attr


def loader(app: Flask):
    """
    The function to load all blueprints based on the ``BLUEPRINTS`` configuration in config.py
    """

    blueprints = app.config.get("BLUEPRINTS", [])
    import_name = get_import_name(app)
    for name in blueprints:
        bp = name + ".routes.init_blueprint"
        routes = name + ".urls.routes"
        try:
            bp: Blueprint = import_attr(import_name + bp)()
            routes = import_attr(import_name + routes)
        except (ImportError, AttributeError):
            bp: Blueprint = import_attr(bp)()
            routes = import_attr(routes)

        for url, view, methods in routes:
            bp.route(url, methods=methods)(view)

        smorest.register_blueprint(bp)



from importlib import import_module

from flask import Flask

from ..helper import get_import_name


def loader(app: Flask):
    """
    Function to load all celery tasks based on ``TASKS`` configuration in config.py.
    """

    dirname = "tasks"
    tasks = app.config.get(dirname.upper(), [])
    import_name = get_import_name(app)
    prefix = dirname + "."
    for name in tasks:
        if not name.startswith(prefix):
            name = prefix + name

        try:
            import_module(import_name + name)
        except ImportError:
            import_module(name.lstrip(prefix))



from functools import wraps
from typing import Callable

from flask import current_app, jsonify
from flask_smorest import Blueprint
from flask_smorest.arguments import ArgumentsMixin
from flask_smorest.response import ResponseMixin
from flask_smorest.pagination import PaginationMixin
from flask_smorest.etag import EtagMixin
from flask_jwt_extended import verify_jwt_in_request
from flask_jwt_extended.utils import get_jwt_claims


def http_code(func: Callable) -> Callable:
    """
    Decorator to generate HTTP status response code automatically
    according to the key ``code`` in json.

    :param func: Your view function.

    """

    @wraps(func)
    def wrapper(*args, **kwds):
        result = func(*args, **kwds)
        code = 200
        if isinstance(result, dict):
            code = result.get("code", 200)

        if not isinstance(result, (tuple, list)):
            result = result, code

        return result

    return wrapper


def api_doc(*args, **kwds) -> Callable:
    """
    See here https://flask-smorest.readthedocs.io/en/latest/api_reference.html#flask_smorest.Blueprint.doc
    """

    wrapper = Blueprint.doc(*args, **kwds)
    return wrapper


def use_kwargs(*args, **kwds) -> Callable:
    """
    See here https://flask-smorest.readthedocs.io/en/latest/arguments.html
    """

    wrapper = ArgumentsMixin().arguments(*args, **kwds)
    return wrapper


def marshal_with(*args, **kwds) -> Callable:
    """
    See here https://flask-smorest.readthedocs.io/en/latest/response.html#
    """

    wrapper = ResponseMixin().alt_response(*args, **kwds)
    return wrapper


def paginate(*args, **kwds) -> Callable:
    """
    See here https://flask-smorest.readthedocs.io/en/latest/pagination.html.
    """

    wrapper = PaginationMixin().paginate(*args, **kwds)
    return wrapper


def etag(schema=None) -> Callable:
    """
    See here https://flask-smorest.readthedocs.io/en/latest/etag.html
    """

    wrapper = EtagMixin().etag(schema)
    return wrapper


def jwt_required(roles={}) -> Callable:
    """
    Decorator to protect views with JWT & user roles.
    """

    def decorated(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwds):
            verify_jwt_in_request()
            claims = get_jwt_claims()
            jwt_roles = claims.get("roles", {})
            for role, permissions in roles.items():
                if not isinstance(permissions, (list, tuple)):
                    permissions = []

                if role in jwt_roles:
                    valid_perms = jwt_roles.get(role, [])
                    for perm in permissions:
                        if perm not in valid_perms:
                            return (
                                jsonify(message="You don't have permission!", code=403),
                                403,
                            )
                else:
                    return (jsonify(message="Role not allowed!", code=403), 403)

            return func(*args, **kwds)

        return wrapper

    return decorated


def authenticate(roles={}) -> Callable:
    """
    Decorator to add jwt view authentication to API Docs.
    Reference: https://github.com/tiangolo/full-stack-flask-couchdb/tree/master/%7B%7Bcookiecutter.project_slug%7D%7D/backend/app/app/api/api_v1.
    """

    def decorated(func: Callable) -> Callable:
        func = api_doc(security=current_app.config.get("API_SECURITY_PARAMS", []))(
            jwt_required(roles)(func)
        )
        return func

    return decorated



import click

from ..generator import g_project


@click.group("app")
def group():
    """
    App generator.
    """


@group.command()
@click.argument("name")
def new(name):
    """
    Create a new app.
    """

    g_project(name, name + ".wsgi")


command = group



from flask_smorest import Api
from flask import Flask


def init_app(app: Flask):
    api = Api(app, spec_kwargs=app.config["API_SPEC_OPTIONS"])
    security_def = app.config["API_SECURITY_DEFINITIONS"]
    for scheme, attrs in security_def.items():
        api.spec.components.security_scheme(scheme, attrs)



import os
from importlib import import_module

from celery import Celery
from flask.cli import load_dotenv

from .scaffold import Scaffold
from .exception import ZemfrogEnvironment
from .helper import get_import_name


def make_celery(app: Scaffold) -> Celery:
    """
    Creating a celery application for flask applications.
    Source: https://flask.palletsprojects.com/en/1.1.x/patterns/celery/#configure
    """

    celery = Celery(
        app.import_name,
        backend=app.config["CELERY_RESULT_BACKEND"],
        broker=app.config["CELERY_BROKER_URL"],
    )
    celery.conf.update(app.config)

    class ContextTask(celery.Task):
        def __call__(self, *args, **kwargs):
            with app.app_context():
                return self.run(*args, **kwargs)

    celery.Task = ContextTask
    return celery


def load_config(app: Scaffold):
    """
    Loads the configuration for your zemfrog application based on the environment
    ``ZEMFROG_ENV``, change your application environment in the file ``.flaskenv``.
    """

    path = os.path.join(app.root_path, ".flaskenv")
    load_dotenv(path)
    env = os.getenv("ZEMFROG_ENV")
    if not env:
        raise ZemfrogEnvironment("environment not found")

    import_name = get_import_name(app)
    app.config.from_object(import_name + "config." + env.capitalize())


def create_app(*args, **kwds) -> Scaffold:
    """
    Functions to build your flask application and load all configurations.

    All parameters will be passed to :py:class:`flask.Flask`.
    """

    app = Scaffold(*args, **kwds)
    import_name = get_import_name(app)
    dirname = "loaders"
    prefix = dirname + "."
    with app.app_context():
        load_config(app)
        api_docs = app.config.get("API_DOCS", True)
        if not api_docs:
            app.config["OPENAPI_URL_PREFIX"] = None

        for name in app.config.get(dirname.upper(), []):
            yourloader = name
            if not name.startswith(prefix):
                yourloader = prefix + yourloader

            try:
                yourloader = import_module(import_name + yourloader)
                loader = getattr(yourloader, "loader")
            except (ImportError, AttributeError):
                yourloader = import_module(name)
                loader = getattr(yourloader, "loader")

            loader(app)

    return app



from datetime import timedelta
from os import path

class Development(object):
    SECRET_KEY = "Your secret key!"
    SQLALCHEMY_DATABASE_URI = "sqlite:///" + path.join(
        path.dirname(__file__), "db.sqlite"
    )
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    JWT_SECRET_KEY = "JWT secret key!"
    JWT_ACCESS_TOKEN_EXPIRES = timedelta(days=7)
    MAIL_PORT = 8025
    MAIL_DEFAULT_SENDER = "admin@localhost.com"
    API_TITLE = "API Docs"
    API_VERSION = "v1"
    API_PREFIX = "/api"
    OPENAPI_VERSION = "3.0.2"
    OPENAPI_JSON_PATH = "openapi.json"
    OPENAPI_URL_PREFIX = "/"
    OPENAPI_REDOC_PATH = "/redoc"
    OPENAPI_REDOC_URL = "https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"
    OPENAPI_SWAGGER_UI_PATH = "/docs"
    OPENAPI_SWAGGER_UI_URL = "https://cdn.jsdelivr.net/npm/swagger-ui-dist/"
    API_SPEC_OPTIONS = {
        'info': {
            'description': 'Powered by [zemfrog](https://github.com/zemfrog/zemfrog) + [flask-smorest](https://flask-smorest.readthedocs.io/en/latest/).',
            'contact' : {
                'email': 'hijriyan23@gmail.com'
            },
            'license': {
                'name': 'MIT License',
                'url': 'https://wikipedia.org/wiki/Licence_MIT'
            }
        }
    }
    API_SECURITY_DEFINITIONS = {
        "Bearer": {
            "type": "oauth2",
            "flows": {
                "password": {
                    "tokenUrl": "/jwt/login"
                }
            }
        }
    }
    API_SECURITY_PARAMS = [{"Bearer": []}]
    DEBUG = True
    {% if main_app -%}
        APPS = []
    {%- endif %}
    EXTENSIONS = [
        "zemfrog.extensions.sqlalchemy",
        "zemfrog.extensions.marshmallow",
        "zemfrog.extensions.migrate",
        "zemfrog.extensions.jwt",
        "zemfrog.extensions.mail",
        "zemfrog.extensions.celery",
        "zemfrog.extensions.smorest",
        "zemfrog.extensions.cors"
    ]
    COMMANDS = [
        "zemfrog.commands.api",
        "zemfrog.commands.blueprint",
        "zemfrog.commands.middleware",
        "zemfrog.commands.command",
        "zemfrog.commands.errorhandler",
        "zemfrog.commands.extension",
        "zemfrog.commands.model",
        "zemfrog.commands.task",
        "zemfrog.commands.user",
        "zemfrog.commands.role",
        "zemfrog.commands.permission",
        "zemfrog.commands.loader",
        "zemfrog.commands.secretkey",
        "zemfrog.commands.context",
        "zemfrog.commands.filter",
        {% if main_app -%}
            "zemfrog.commands.app"
        {%- endif %}
    ]
    BLUEPRINTS = []
    STATICFILES = []
    MIDDLEWARES = []
    APIS = []
    ERROR_HANDLERS = {}
    TASKS = ["zemfrog.tasks"]
    CONTEXT_PROCESSORS = []
    JINJA_FILTERS = []
    API_DOCS = True
    CREATE_DB = True
    USER_MODEL = "models.user.User"
    ROLE_MODEL = "models.user.Role"
    PERMISSION_MODEL = "models.user.Permission"
    LOG_MODEL = "models.user.Log"
    LOADERS = [
        "zemfrog.loaders.extension",
        "zemfrog.loaders.staticfile",
        "zemfrog.loaders.model",
        "zemfrog.loaders.blueprint",
        "zemfrog.loaders.middleware",
        "zemfrog.loaders.api",
        "zemfrog.loaders.error_handler",
        "zemfrog.loaders.command",
        "zemfrog.loaders.task",
        "zemfrog.loaders.context_processor",
        "zemfrog.loaders.jinja_filter",
        "zemfrog.loaders.multiapp",
    ]
    CELERY_RESULT_BACKEND = "redis://127.0.0.1:6379"
    CELERY_BROKER_URL = CELERY_RESULT_BACKEND


class Production(Development):
    DEBUG = False
    JWT_COOKIE_SECURE = True


class Testing(Development):
    TESTING = True



"""
Reference:
* https://github.com/prompt-toolkit/python-prompt-toolkit/blob/master/examples/prompts/fancy-zsh-prompt.py
* https://github.com/prompt-toolkit/python-prompt-toolkit/blob/master/examples/prompts/auto-completion/nested-autocompletion.py

"""

import datetime
import os

from click import Group
from flask import Flask
from prompt_toolkit import prompt
from prompt_toolkit.application import get_app
from prompt_toolkit.completion import NestedCompleter
from prompt_toolkit.formatted_text import (
    HTML,
    fragment_list_width,
    merge_formatted_text,
    to_formatted_text,
)
from prompt_toolkit.styles import Style

style = Style.from_dict(
    {
        "import_name": "#aaaaaa italic",
        "path": "#ffffff bold",
        "env": "bg:#666666",
        "left-part": "bg:#444444",
        "right-part": "bg:#444444",
        "padding": "bg:#444444",
    }
)


def get_info_app(app: Flask):
    left_part = HTML(
        "<left-part>" " <import_name>%s</import_name> " "<path>%s</path>" "</left-part>"
    ) % (app.import_name, app.root_path)
    right_part = HTML(
        "<right-part> " " <env>%s</env> " " <time>%s</time> " "</right-part>"
    ) % (os.environ["ZEMFROG_ENV"], datetime.datetime.now().isoformat())

    used_width = sum(
        [
            fragment_list_width(to_formatted_text(left_part)),
            fragment_list_width(to_formatted_text(right_part)),
        ]
    )

    total_width = get_app().output.get_size().columns
    padding_size = total_width - used_width

    padding = HTML("<padding>%s</padding>") % (" " * padding_size,)
    return left_part, padding, right_part, "\n"


def get_prompt(app: Flask):
    """
    Build the prompt dynamically every time its rendered.
    """

    info_app = []
    if getattr(app, "_show_info_app", True):
        info_app = get_info_app(app)

    return merge_formatted_text([*info_app, "$ "])


def get_commands(cli: Group) -> dict:
    commands = {}
    for name, cmd in cli.commands.items():
        value = None
        if isinstance(cmd, Group):
            value = get_commands(cmd)
        commands[name] = value
    return commands


def get_auto_complete(cli: Group) -> NestedCompleter:
    commands = {"flask": get_commands(cli), "info": None}
    completer = NestedCompleter.from_nested_dict(commands)
    return completer


def build_repl(app: Flask):
    while True:
        try:
            answer = prompt(
                lambda: get_prompt(app),
                style=style,
                completer=get_auto_complete(app.cli),
                refresh_interval=1,
            )
            if answer == "info":
                print("* Import name: " + app.import_name)
                print("* Root path: " + app.root_path)
                print("* Environment: " + os.environ["ZEMFROG_ENV"])
            else:
                os.system(answer)

            app._show_info_app = False

        except (EOFError, KeyboardInterrupt):
            print("Bye bye!")
            break



from importlib import import_module

from flask import Flask
from flask_smorest import Blueprint
from ..globals import smorest
from ..helper import get_import_name


def loader(app: Flask):
    """
    A function to load all of your API resources to flask based on the ``APIS`` configuration in config.py.
    """

    apis = app.config.get("APIS", [])
    import_name = get_import_name(app)
    api_prefix = app.config.get("API_PREFIX", "/api")
    prefix = "apis."
    for name in apis:
        res = name
        if not name.startswith(prefix):
            res = prefix + res

        try:
            res = import_module(import_name + res)
        except ImportError:
            res = import_module(name.lstrip(prefix))

        tag = res.tag
        description = res.description
        url_prefix = res.url_prefix
        bp = Blueprint(
            tag, __name__, url_prefix=api_prefix + url_prefix, description=description
        )
        routes = res.routes
        for detail in routes:
            url, view, methods = detail
            bp.route(url, methods=methods)(view)

        smorest.register_blueprint(bp)



import os
import subprocess
from typing import Dict, List

import click
from flask import Flask
from werkzeug.middleware.dispatcher import DispatcherMiddleware

from ..helper import import_attr
from ..repl import build_repl


def loader(app: Flask):
    """
    Load all applications and combine them together using ``DispatcherMiddleware``.
    """

    apps: List[Dict] = app.config.get("APPS", [])
    mounts = {}
    for a in apps:
        if not isinstance(a, dict):
            a = {"name": a}

        name = a["name"]
        path = a.get("path", "/" + name)
        help = a.get("help", f"{name.title()} App")
        yourapp: Flask = import_attr(name + ".wsgi.app")

        @click.command(name, context_settings=dict(ignore_unknown_options=True))
        @click.option("-r", "--repl", is_flag=True, help="Activates REPL mode.")
        @click.argument("args", nargs=-1, type=click.UNPROCESSED)
        def cli(repl, args):
            os.environ["FLASK_APP"] = yourapp.import_name
            if repl:
                build_repl(yourapp)
            else:
                subprocess.call(["flask"] + list(args))

        cli.help = help
        app.cli.add_command(cli)
        mounts[path] = yourapp

    app.wsgi_app = DispatcherMiddleware(app.wsgi_app, mounts)



import os

from flask.app import Flask
from flask_migrate import Migrate

from ..globals import db


def init_app(app: Flask):
    migrate = Migrate(db=db)
    directory = os.path.join(
        app.root_path, "migrations/" + os.getenv("ZEMFROG_ENV", "development").lower()
    )
    migrate.init_app(app, directory=directory)



from flask import jsonify
from werkzeug.exceptions import HTTPException


def handler(error: HTTPException):
    code = error.code
    return jsonify(message="error occurred %r" % error.description, code=code), code
